/*
  Copyright (C) 2020 Alessandro Languasco & Luca Righi
*/
#include "compute.h"

#define REAL 0
#define IMAG 1

long double Eulergamma = 0.57721566490153286060651209008240243104;
long double pi = 3.1415926535897932384626433832795028841971693993751;

/* for odd/even results */
struct result {
  long double res;
  long double max;
  long double min;
};

/* mmap.c */
void * mapRO_File2Array(const char *filename, int* fd, size_t arraySize, size_t typeSize);
long double *mapRWCreate_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize);
long double *mapRW_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize);
void unmap_File2Array(void * mmappedData, size_t filesize, int fd);
char * mallocFilename(const char * basename, const char* suffix);
void * mkmmapcopy(void * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix, const char mode, int *fd);
fftwl_complex * mapfile(const char * name, const char * basename, long int size, int * fd);
void  print_time(const char * message, double startplan, double endplan);
/* s.c */
long int lpow_mod(long int base, long int exp, long int mod);
long double s(long double aq, int precision, int verbose);

struct result readResult(const char * basename, const char * name) {
  const char * funcname = "readResult()";
  
  struct result * r = NULL;
  if ((r = malloc(sizeof(struct result))) == NULL) {
    fprintf(stderr, "%s: unable to allocate result. Giving up\n", funcname);
    exit(1);
  }
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "r")) == NULL) {
    fprintf(stderr, "%s: can't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }
  
  if (fscanf(f, "%Lf\n", &(r->res)) != 1) {
    fprintf(stderr, "%s: unable to read 'res' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf\n", &(r->max)) != 1) {
    fprintf(stderr, "%s: unable to read 'max' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf", &(r->min)) != 1) {
    fprintf(stderr, "%s: unable to read 'min' from %s: %s\n", funcname, filename, strerror(errno));
  }

  fclose(f);

  printf("%s: readed %Lf, %Lf, %Lf from %s\n", funcname, r->res, r->max, r->min, filename);
  
  return *r;
}

void writeResult(struct result r, const char * basename, const char * name) {

  const char * funcname = "writeResult()";
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "w")) == NULL) {
    fprintf(stderr, "%s: san't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }

  fprintf(f, "%Lf\n", r.res);
  fprintf(f, "%Lf\n", r.max);
  fprintf(f, "%Lf\n", r.min);

  fclose(f);

  return;
}

fftwl_complex * mkfftin(ptrdiff_t nhalf, const char *basename, int *fd_fftin) {

  char * funcname = "mkfftin()";
  char * fftinFilename = mallocFilename(basename, "-IN.ld");

  fftwl_complex * fftin = (fftwl_complex *) mapRWCreate_File2Array(fftinFilename, fd_fftin, nhalf, sizeof(fftwl_complex));
  
  if (fftin == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftinFilename);
    return NULL;
  }

  return fftin;
}

fftwl_complex * mkfftout(ptrdiff_t nhalf, const char *basename, int *fd_fftout) {

  char * funcname = "mkfftout()";
  char * fftoutFilename = mallocFilename(basename, "-OUT.ld");

  fftwl_complex * fftout = (fftwl_complex *) mapRWCreate_File2Array(fftoutFilename, fd_fftout, nhalf, sizeof(fftwl_complex));
  if (fftout == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftoutFilename);
    return 0;
  }

  return fftout;
}

fftwl_plan mkplan(fftwl_complex *fftin, fftwl_complex *fftout, ptrdiff_t nhalf) {

  clock_t startplan = clock();
  
  if (fftwl_import_wisdom_from_filename("wisdom-inplace-fftwl-estimate.txt")) {
    fprintf(stderr,  "*** Reading wisdom-inplace-fftwl-estimate \n");
  } else {
    fprintf(stderr,  "*** Generating wisdom-inplace-fftwl-estimate \n");
  }

  int rank=1;
  fftw_iodim64 * dims = malloc(rank * sizeof(fftw_iodim64));
  dims[0].n= nhalf;   /* length of the trasform */
  dims[0].is=1;
  dims[0].os=1; 
  int howmany_rank=1; 
  fftw_iodim64 * howmany_dims = malloc(howmany_rank * sizeof(fftw_iodim64)); 
  howmany_dims[0].n=1;
  howmany_dims[0].is=1;
  howmany_dims[0].os=1;
  
  fftwl_plan plan = fftwl_plan_guru64_dft(rank, /*rank=1 means monodimensional vector*/
              dims, /* nhalf */
              howmany_rank,
              howmany_dims,
              fftin,   /*input for INPLACE transforms*/ 
              fftout,  /*output; for INPLACE transforms*/ 
              FFTW_FORWARD, /* -1 because it is the sign of FORWARD transform */
              FFTW_ESTIMATE); /*flags*/
              
  if (plan == NULL) { 
    fprintf(stderr,"plan creation failed\n");
    return 0;
  }
    
  /*  saving  a  generated  wisdom  for  future  runs  */
  if (!fftwl_import_wisdom_from_filename("wisdom-inplace-fftwl-estimate.txt")) {
    fprintf(stderr,  "***  Saving wisdom-inplace-fftwl-estimate.txt \n");
    fftwl_export_wisdom_to_filename("wisdom-inplace-fftwl-estimate.txt"); 
  }
  
  clock_t endplan = clock(); 

  print_time("Plan generation", startplan, endplan);

  return plan;
}

fftwl_complex * mkLGodd(fftwl_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_LGodd) {

  long double pi = 3.1415926535897932384626433832795028841971693993751;
  long double sinvalue;
  long double cosvalue;
  long double logpi = logl(pi);
  long double twist = 0.0L;
  long double angle = 0.0L; 
  long double aux;
  long double d;
  long int a;
  ptrdiff_t k;
  
  /* 
   * STEP: mkLGodd
   *  need: fftin mmapped array
   *  provide: file <basename>-LGodd.ld
   */
  
  clock_t startclock = clock();  /* start clock */

  angle = (long double) pi / nhalf;
  twist = 0.0L; 
  a = 1; 

  for (k = 0; k <nhalf ; k++){   
     cosvalue = cosl(twist);
     sinvalue = sinl(twist);  
     d = (long double) a/q ;
     aux = (2.0) * lgammal( d ) + logl( sinl(pi * d)) - logpi ;
     fftin[k][REAL] = aux * cosvalue;  
     fftin[k][IMAG] = (-1.0L) * aux * sinvalue; 
     a = (a * primroot) % q;  
     twist = twist + angle;
  }

  clock_t endclock = clock();  /* end clock */  
  print_time("LNGammaOdd computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwl_complex * lgodd = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_LGODD, 'r', fd_LGodd);

  return lgodd;

}

fftwl_complex * fftLGodd(fftwl_complex *fftin, fftwl_complex *fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_LGoddfft) {
  
  /* 
   * STEP: fftLGodd
   *  need: fftin mmapped array with LGodd data
   *  provide: fft of fftin in array fftout
   *  provide: copy of fftout in file <basename>-LGoddfft.ld if basename != NULL
   */

  clock_t startfft = clock(); /* start clock */
  fftwl_execute(plan); 
  clock_t endfft = clock(); /* stop clock */
  print_time("FFT(LGodd)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwl_complex * lgoddfft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTLGODD, 'r', fd_LGoddfft);

  return lgoddfft;

}

fftwl_complex * mkAK(fftwl_complex *fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_AK) {

  long double pi = 3.1415926535897932384626433832795028841971693993751;
  long double sinvalue;
  long double cosvalue;
  long double twist = 0.0L;
  long double angle = 0.0L; 
  long double c;
  ptrdiff_t a;
  ptrdiff_t k;

  /* 
   * STEP: mkAK
   *  need: fftin mmapped array available for writing
   *  provide: input data AK (in fftin) to compute <basename>-AKfft.ld
   */
  
  clock_t startclock = clock();  /* start clock */  
  angle = (long double) pi/nhalf;
  twist = 0.0L; 
  a = 1; 
  for (k = 0; k < nhalf ; k++){  
     c = (long double) (2*a-q); /*g^k -R g^(k+m) \equiv g^k  -R (q-1)g^k \equiv g^k -R (-g^k) \equiv g^k -R (q -g^k) = 2*g^k -R q */
     cosvalue = cosl(twist);
     sinvalue = - sinl(twist);  /* depends on the sign = -1 of the transform */
     fftin[k][REAL] = c * cosvalue;
     fftin[k][IMAG] = c * sinvalue;
     a = (a * primroot) % q;
     twist += angle;
  }
  clock_t endclock = clock();  /* end clock */
  print_time("AK computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwl_complex * ak = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_AK, 'r', fd_AK);

  return ak;

}

fftwl_complex * fftAK(fftwl_complex *fftin, fftwl_complex *fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_AKfft) {
  
  /* STEP: fftAK 
   *  need: fftin mmapped array with ak start data 
   *  provide: fft in fftout, file <basename>-AKoddfft.ld if basename != NULL
   */

  clock_t startfft = clock();
  fftwl_execute(plan);
  clock_t endfft  = clock();
  print_time("FFT(AK)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  fftwl_complex * akfft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTAK, 'r', fd_AKfft);

  return akfft;

}

struct result oddResult(fftwl_complex *lngammaodd_fft, fftwl_complex *ak_fft, ptrdiff_t nhalf, long int q) {
   
  struct result myresult;
  
  /* Computing EK(q) - EK^+(q) */
  /* Computing min_{chi\ne chi_0; chi odd} \vert L'/L(1,\chi) \vert =: mq_odd */
  /* Computing max_{chi\ne chi_0; chi odd} \vert L'/L(1,\chi) \vert =: Mq_odd */ 
  
  printf("%s", "Computing the sum and minmax over odd characters"); 
  fflush(stdout);
  
  clock_t startcompodd = clock(); /* start clock */

  long double pi = 3.1415926535897932384626433832795028841971693993751;
  long double oddresult = 0;
  long double res;
  long double Minimumoddresult;
  long double Maximumoddresult;
  long double Minimum = LDBL_MAX;/* odd characters */
  long double Maximum = 0; 
  long double num     = 0;
  long double num1    = 0;
  long double den     = 0;
  long double quot    = 0;   
  long double frac    = 0;   
  long double frac1   = 0;   
  long double corr    = logl(2*pi)+Eulergamma;  
  res = 0;
  oddresult = 0;

  /* computes |(a+ib)/(c+id)|= |a+ib|/|c+id| and the real part of (a+ib)/(c+id); the imaginary part is 0 by def */
  for (long int i=0;  i < nhalf; ++i) { 
    num = lngammaodd_fft[i][REAL] * ak_fft[i][REAL] + lngammaodd_fft[i][IMAG] * ak_fft[i][IMAG];
    den = ak_fft[i][REAL] * ak_fft[i][REAL]+ ak_fft[i][IMAG] * ak_fft[i][IMAG];
    res = res + num/den;
    num1 = lngammaodd_fft[i][IMAG] * ak_fft[i][REAL] - lngammaodd_fft[i][REAL] * ak_fft[i][IMAG];
    frac = corr + q * num / den;
    frac1 = q * num1 / den;
    quot = frac * frac + frac1 * frac1;
    if (quot < Minimum) { Minimum = quot; };
    if (quot > Maximum) { Maximum = quot; }; 
  }
  
  Minimumoddresult = sqrtl(Minimum); /* min for odd characters */
  Maximumoddresult = sqrtl(Maximum); /* max for odd characters */
  oddresult = q * res; /* correction factors for odd characters */

  myresult.res = oddresult;
  myresult.max = Maximumoddresult;
  myresult.min = Minimumoddresult;
  
  clock_t endcompodd = clock(); /* stop clock */
  printf("\n *** First: oddresult = %32.30Lf, (res = %24.28Lf)\n", oddresult, res);
  print_time("OddResult computation", startcompodd, endcompodd);
  
  fflush(stdout);

  return myresult;
}

fftwl_complex * mkS(fftwl_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, int precision, const char * basename, int * fd_S) {
  
  /* STEP: S 
   *  need: fftin array available for writing
   *  provide: S data in fftin to compute <basename>-Sfft.ld
   *  provide: 
   */
  
  clock_t startcomputeS = clock();
  
  /* inserting the null imaginary parts */
  /*long int s_a = lpow_mod(primroot, 0, q); */  /* Ale: qui si puo' mettere s_a=1 */
  long int s_a = 1; 
  long double s_aq = 0.0;
  for  (long int k = 0; k < nhalf; k++) {  /* Ale: qui si dovrebbe mettere nhalf invece di nhalf+2 */
    s_aq = (long double) ((long double) s_a / (long double) q);
    fftin[k][REAL] = s(s_aq, precision, 0); 
    fftin[k][IMAG] = 0.0L;
    s_a = (s_a * primroot) % q;
  }
  
  clock_t endcomputeS = clock();
  fflush(stdout);
  print_time("S computation", startcomputeS, endcomputeS);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  fftwl_complex * S = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_S, 'r', fd_S);

  return S;
}

fftwl_complex * fftS(fftwl_complex * fftin, fftwl_complex * fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_Sfft) {
  
  /* STEP: fftS 
   *  need: fftin mmapped array with S data 
   *  provide: file <basename>-Sfft.ld (by copying fftout)
   */
 
  clock_t startfft  = clock();
  fftwl_execute(plan);
  clock_t endfft = clock();
  print_time("Sfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwl_complex * S_fft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTS, 'r', fd_Sfft);

  return S_fft;
}

fftwl_complex * mkLGeven(fftwl_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char *basename, int * fd_LGeven) {
  
  /*
   * STEP: mkLGeven 
   *  need: fftin mmapped array
   *  provide: file <basename>-LGeven.ld
   *
   */
  
  long double pi = 3.1415926535897932384626433832795028841971693993751;
  long double logpi = logl(pi);
  long double d;
  long int a;  
  long int k;

  clock_t startcomp = clock();
  a = 1; 
  for (k = 0; k < nhalf; k++){  
     d = (long double) a/q ;
     fftin[k][REAL]= logpi - logl(sinl(pi*d));  /* values of log(gamma(a_k/q) via the reflection formula*/
     fftin[k][IMAG] = 0.0L;  
     a = (a * primroot) % q;
  }
  clock_t endcomp = clock();
  print_time("mkLGeven", startcomp, endcomp);
  fflush(stdout);
  
  if (basename == NULL) {
    /* copy of the file not requested, output in fftin */
    return NULL;
  }

  fftwl_complex * LGeven = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_LGEVEN, 'r', fd_LGeven);

  return LGeven;

}

fftwl_complex * fftLGeven(fftwl_complex *fftin, fftwl_complex *fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_LGevenfft) {
  
  /* STEP: fftLGeven 
   *  need: fftin mmapped array with LGeven data
   *  provide: file <basename>-LGevenfft.ld (by renaming fftout)
   */
  
  clock_t startfft = clock();
  fftwl_execute(plan);
  clock_t endfft = clock();
  print_time("LGevenfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwl_complex * LGevenfft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTLGEVEN, 'r', fd_LGevenfft);

  return LGevenfft;

}

struct result evenResult(fftwl_complex *lngamma_evenfft, fftwl_complex *S_fft, ptrdiff_t nhalf, long int q) {

  /* STEP: evenResult */
  
  /* Computing EK^+(q) */
  /* Computing min_{chi\ne chi_0; chi even} \vert L'/L(1,\chi) \vert =: mq_even */
  /* Computing max_{chi\ne chi_0; chi even} \vert L'/L(1,\chi) \vert =: Mq_even */
  /* sum starts from i=1 to avoid the trivial character contribution */

  struct result myresult;
  
  long double evenresult = 0;
  long double res = 0;
  long double Minimumevenresult = 0;
  long double Maximumevenresult = 0;
  long double Minimum = LDBL_MAX; /* even characters */
  long double Maximum = 0; 
  long double num     = 0;
  long double num1    = 0;
  long double den     = 0;
  long double quot    = 0;   
  long double frac    = 0;   
  long double frac1   = 0;   
  long double corr    = logl(2*pi)+Eulergamma;
  /* ptrdiff_t nquarter = nhalf / 2; */
  /* ptrdiff_t m; */

  res = 0;
  evenresult = 0;

  printf("Computing the sum and minmax over even characters ..."); 
  clock_t startcompeven = clock();

  /* sum starts from i=1 to avoid the trivial character contribution */
  /* il for da riga 560 e' scritto cosi' perche' prima erano trasformate di successioni reali con simmetria;
     ora invece sono complesse; quindi si potrebbe semplificare mettedo come nel caso odd ; ossia come segue */
     
  for (long int i=1;  i < nhalf; ++i) { 
      num = S_fft[i][REAL] * lngamma_evenfft[i][REAL] +  S_fft[i][IMAG]*lngamma_evenfft[i][IMAG];
      den = lngamma_evenfft[i][REAL] *lngamma_evenfft[i][REAL] + lngamma_evenfft[i][IMAG]*lngamma_evenfft[i][IMAG];
      res = res + num / den;
      num1 = S_fft[i][IMAG] * lngamma_evenfft[i][REAL] - S_fft[i][REAL]*lngamma_evenfft[i][IMAG];
      frac = corr + num/(2*den);
      frac1 = num1 / (2*den);
      quot = frac * frac + frac1 * frac1; 
      if (quot < Minimum) { Minimum = quot; }
      if (quot > Maximum) { Maximum = quot; } 
 
  }   

  /* sum starts from i=1 to avoid the trivial character contribution */
  /* lngamma_evenfft contains the DFT of lngamma_even; the output has simmetries 'cause in fact the imput sequence is real */
  /* lngamma_evenfft[i][REAL]=lngamma_evenfft[nhalf-i][REAL];  lngamma_evenfft[i][IMAG] = -lngamma_evenfft[nhalf-i][IMAG] */
  /* S_fft contains the DFT of S; the output has simmetries 'cause in fact the imput sequence is real */
  /* S_fft[i][REAL]=S_fft[nhalf-i][REAL];  S_fft[i][IMAG] = -S_fft[nhalf-i][IMAG] */
  /*   SO THEY HAVE THE SAME ABSOLUTE VALUES  */
  /* computes |(a+ib)/(c+id)|= |a+ib|/|c+id| and the real part of (a+ib)/(c+id); the imaginary part is 0 by def */
  
  /*
  for (long int i=1;  i <= nhalf-1; ++i) { 
    if ( i <= nquarter ) {
      num = S_fft[i][REAL] * lngamma_evenfft[i][REAL] +  S_fft[i][IMAG]*lngamma_evenfft[i][IMAG];
      den = lngamma_evenfft[i][REAL] *lngamma_evenfft[i][REAL] + lngamma_evenfft[i][IMAG]*lngamma_evenfft[i][IMAG];
      res = res + num / den;
      num1 = S_fft[i][IMAG] * lngamma_evenfft[i][REAL] - S_fft[i][REAL]*lngamma_evenfft[i][IMAG];
      frac = corr + num/(2*den);
      frac1 = num1 / (2*den);
      quot = frac * frac + frac1 * frac1; 
      if (quot < Minimum) { Minimum = quot; }
      if (quot > Maximum) { Maximum = quot; } 
    } else {
      m = nhalf-i;
      num = S_fft[m][REAL] * lngamma_evenfft[m][REAL] +  S_fft[m][IMAG]*lngamma_evenfft[m][IMAG];
      den = lngamma_evenfft[m][REAL] * lngamma_evenfft[m][REAL] + lngamma_evenfft[m][IMAG]*lngamma_evenfft[m][IMAG];
      res = res + num / den; 
      num1 = S_fft[m][IMAG] * lngamma_evenfft[m][REAL] - S_fft[m][REAL]*lngamma_evenfft[m][IMAG];
      frac = corr + num/(2 * den);
      frac1 = num1 / (2 * den);
      quot = frac * frac + frac1 * frac1; 
      if (quot < Minimum) { Minimum = quot; }
      if (quot > Maximum) { Maximum = quot; } 
    }
  }
  */
  
  Minimumevenresult = sqrtl(Minimum); /* min for even characters */
  Maximumevenresult = sqrtl(Maximum); /* max for even characters */
  evenresult = res/2.0;

  myresult.res = evenresult;
  myresult.max = Maximumevenresult;
  myresult.min = Minimumevenresult;
  
  clock_t endcompeven = clock();

  printf("\n *** Second: evenresult = %32.30Lf, (res = %24.28Lf)\n", evenresult, res);
  print_time("EvenResult computation", startcompeven, endcompeven);
  
  fflush(stdout);

  return myresult;
}


int printResults(long int q, ptrdiff_t nhalf, struct result oddresult, struct result evenresult) {

  /** RESULTS **/
  printf("*** RESULTS:\n\n");
  fflush(stdout);
  long double EK = Eulergamma + evenresult.res + oddresult.res + (Eulergamma + logl(2*pi)) *(q-2);
  long double EK_plus = nhalf*Eulergamma + (nhalf-1)* logl(2*pi) + evenresult.res;
  long double EK_diff = oddresult.res + (Eulergamma + log(2*pi)) * nhalf;
  
  if (EK < 0 || EK_plus < 0 ) {
    FILE * negativeEKfile;
    negativeEKfile = fopen ("../negativeEK.txt","a");
    fprintf (negativeEKfile, "q = %td, ", q);
    fprintf (negativeEKfile, "EK = %32.30Lf, ", EK); 
    fprintf (negativeEKfile, "EK+ = %32.30Lf, ", EK_plus); 
    fprintf (negativeEKfile, "EK-EK+ = %32.30Lf\n", EK_diff); 
    fclose (negativeEKfile);
  };

  printf("%s\n", "--------------");
  /* print EK_q */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", EK);
  
  /* print EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")^+ = ");
  printf("%32.30Lf\n", EK_plus); 
  
  /* print EK_q- EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ") - ");
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")^+ = ");
  printf("%32.30Lf\n", EK_diff); 

  printf("%s\n", "--------------");
  
  /* print min_q */
  printf("%s", "min_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.min); 

  /* print min_q */
  printf("%s", "min_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.min);  

  
  /* print min_q */
  printf("%s", "min_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fminl(evenresult.min, oddresult.min));    
  
  printf("%s\n", "--------------");
  
  /* print max_q */
  printf("%s", "max_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.max); 

  /* print max_q */
  printf("%s", "max_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.max);  

  
  /* print min_q */
  printf("%s", "max_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fmaxl(evenresult.max, oddresult.max));   
  
  printf("%s\n", "--------------");  

  return 1;
}

int dostep(const char * step, const char * steps) {
  if (strcasestr(steps, step) != NULL) {
    return 1;
  }
  return 0;
}

int runComputation(long int q, long int primroot, int precision, const char * basename, const char * steps) {

  ptrdiff_t NUM_POINTS = q-1;
  ptrdiff_t nhalf = NUM_POINTS / 2;

  int fd_fftin;
  int fd_fftout;
  int fd_LGodd;
  int fd_LGoddfft;
  int fd_AK;
  int fd_AKfft;
  int fd_S;
  int fd_Sfft;
  int fd_LGeven;
  int fd_LGevenfft;
  
  fftwl_complex * fftin  = mkfftin(nhalf, basename, &fd_fftin);
  fftwl_complex * fftout = mkfftout(nhalf, basename, &fd_fftout);
  fftwl_complex * lgodd = NULL;
  fftwl_complex * lgoddfft = NULL;
  fftwl_complex * ak = NULL;
  fftwl_complex * akfft = NULL;
  fftwl_complex * S = NULL;
  fftwl_complex * Sfft = NULL;
  fftwl_complex * lgeven = NULL;
  fftwl_complex * lgevenfft = NULL;
  
  struct result oddresult;
  struct result evenresult;
  
  int doall = 0; /* make every step */
  
  if (fftin == NULL || fftout == NULL)
    return 0;

  fftwl_plan plan = mkplan(fftin, fftout, nhalf);

  if (plan == NULL)
    return 0;

  if (dostep(ALL, steps)) {
    doall = 1;
  }
  
  if (doall || dostep(ODDRESULT, steps)) {
  
    if (doall || dostep(MKLGODD, steps)) {
      lgodd     = mkLGodd(fftin, nhalf, q, primroot, basename, &fd_LGodd);       /* qui non servirebbe salvare */
    } else {
      lgodd     = mapfile(N_LGODD, basename, nhalf, &fd_LGodd);
    }
    
    if (doall || dostep(FFTLGODD, steps)) {
      lgoddfft  = fftLGodd(fftin, fftout, plan, nhalf, basename, &fd_LGoddfft);
    } else {
      lgoddfft  = mapfile(N_FFTLGODD, basename, nhalf, &fd_LGoddfft);
    }
    
    if (doall || dostep(MKAK, steps)) {
      ak        = mkAK(fftin, nhalf, q, primroot, basename, &fd_AK);             /* qui non servirebbe salvare */
    } else {
      ak        = mapfile(N_AK, basename, nhalf, &fd_AK);
    }
    
    if (doall || dostep(FFTAK, steps)) {
      akfft     = fftAK(fftin, fftout, plan, nhalf, basename, &fd_AKfft);        /* qui non servirebbe salvare */
    } else {
      akfft     = mapfile(N_FFTAK, basename, nhalf, &fd_AKfft);
    }
    
    oddresult   = oddResult(lgoddfft, akfft, nhalf, q);

    writeResult(oddresult, basename, N_ODDRESULT);
    
    unmap_File2Array(lgodd, nhalf, fd_LGodd);
    unmap_File2Array(lgoddfft, nhalf, fd_LGoddfft);
    unmap_File2Array(ak, nhalf, fd_AK);
    unmap_File2Array(akfft, nhalf, fd_AKfft);

  } else {

    oddresult = readResult(basename, N_ODDRESULT);

  }


  if (doall || dostep(EVENRESULT, steps)) {
  
    if (doall || dostep(MKS, steps)) {
      S         = mkS(fftin, nhalf, q, primroot, precision, basename, &fd_S);    /* qui non servirebbe salvare */
    } else {
      S         = mapfile(N_S, basename, nhalf, &fd_S);
    }
    
    if (doall || dostep(FFTS, steps)) {
      Sfft      = fftS(fftin, fftout, plan, nhalf, basename, &fd_Sfft);
    } else {
      Sfft      = mapfile(N_FFTS, basename, nhalf, &fd_Sfft);
    }
    
    if (doall || dostep(MKLGEVEN, steps)) {
      lgeven    = mkLGeven(fftin, nhalf, q, primroot, basename, &fd_LGeven);      /* qui non servirebbe salvare */
    } else {
      lgeven    = mapfile(N_LGEVEN, basename, nhalf, &fd_LGeven);
    }
    
    if (doall || dostep(FFTLGEVEN, steps)) {
      lgevenfft = fftLGeven(fftin, fftout, plan, nhalf, basename, &fd_LGevenfft); /* qui non servirebbe salvare */
    } else {
      lgevenfft = mapfile(N_FFTLGEVEN, basename, nhalf, &fd_LGevenfft);
    }
    
    evenresult  = evenResult(lgevenfft, Sfft, nhalf, q);

    writeResult(evenresult, basename, N_EVENRESULT);
    
    unmap_File2Array(S, nhalf, fd_S);
    unmap_File2Array(Sfft, nhalf, fd_Sfft);
    unmap_File2Array(lgeven, nhalf, fd_LGeven);
    unmap_File2Array(lgevenfft, nhalf, fd_LGevenfft);
    
  } else {

    evenresult = readResult(basename, N_EVENRESULT);

  }

  printResults(q, nhalf, oddresult, evenresult);

  fftwl_destroy_plan(plan);
  fftwl_cleanup();

  return 0;
}

