/*
  Copyright (C) 2020 Alessandro Languasco & Luca Righi
*/
#include "computeq.h"
#include <quadmath.h>

#define REAL 0
#define IMAG 1

__float128 Eulergamma = 0.57721566490153286060651209008240243104;
__float128 pi = 3.1415926535897932384626433832795028841971693993751;

/* for odd/even results */
struct result {
  __float128 res;
  __float128 max;
  __float128 min; 
};

/* mmap.c */
void * mapRO_File2Array(const char *filename, int* fd, size_t arraySize, size_t typeSize);
__float128 *mapRWCreate_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize);
__float128 *mapRW_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize);
void unmap_File2Array(void * mmappedData, size_t filesize, int fd);
char * mallocFilename(const char * basename, const char* suffix);
void * mkmmapcopy(void * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix, const char mode, int *fd);
fftwq_complex * mapfile(const char * name, const char * basename, long int size, int * fd);
void  print_time(const char * message, double startplan, double endplan);
/* s.c */
long int lpow_mod(long int base, long int exp, long int mod);
__float128 s(__float128 aq, int precision, int verbose);

struct result readResult(const char * basename, const char * name) {
  const char * funcname = "readResult()";
  
  struct result * r = NULL;
  if ((r = malloc(sizeof(struct result))) == NULL) {
    fprintf(stderr, "%s: unable to allocate result. Giving up\n", funcname);
    exit(1);
  }
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "r")) == NULL) {
    fprintf(stderr, "%s: can't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }
  
  if (fscanf(f, "%Lf\n", &(r->res)) != 1) {
    fprintf(stderr, "%s: unable to read 'res' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf\n", &(r->max)) != 1) {
    fprintf(stderr, "%s: unable to read 'max' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf", &(r->min)) != 1) {
    fprintf(stderr, "%s: unable to read 'min' from %s: %s\n", funcname, filename, strerror(errno));
  }

  fclose(f);

  printf("%s: read %Lf, %Lf, %Lf from %s\n", funcname, r->res, r->max, r->min, filename);
  
  return *r;
}

void writeResult(struct result r, const char * basename, const char * name) {

  const char * funcname = "writeResult()";
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "w")) == NULL) {
    fprintf(stderr, "%s: san't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }

  fprintf(f, "%Lf\n", r.res);
  fprintf(f, "%Lf\n", r.max);
  fprintf(f, "%Lf\n", r.min);

  fclose(f);

  return;
}

fftwq_complex * mkfftin(ptrdiff_t nhalf, const char *basename, int *fd_fftin) {

  char * funcname = "mkfftin()";
  char * fftinFilename = mallocFilename(basename, "-IN.qd");

  fftwq_complex * fftin = (fftwq_complex *) mapRWCreate_File2Array(fftinFilename, fd_fftin, nhalf, sizeof(fftwq_complex));
  
  if (fftin == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftinFilename);
    return NULL;
  }

  return fftin;
}

fftwq_complex * mkfftout(ptrdiff_t nhalf, const char *basename, int *fd_fftout) {

  char * funcname = "mkfftout()";
  char * fftoutFilename = mallocFilename(basename, "-OUT.qd");

  fftwq_complex * fftout = (fftwq_complex *) mapRWCreate_File2Array(fftoutFilename, fd_fftout, nhalf, sizeof(fftwq_complex));
  if (fftout == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftoutFilename);
    return 0;
  }

  return fftout;
}

fftwq_plan mkplan(fftwq_complex *fftin, fftwq_complex *fftout, ptrdiff_t nhalf) {

  clock_t startplan = clock();
  
  if (fftwq_import_wisdom_from_filename("wisdom-inplace-fftwq-estimate.txt")) {
    fprintf(stderr,  "*** Reading wisdom-inplace-fftwq-estimate \n");
  } else {
    fprintf(stderr,  "*** Generating wisdom-inplace-fftwq-estimate \n");
  }

  int rank=1;
  fftw_iodim64 * dims = malloc(rank * sizeof(fftw_iodim64));
  dims[0].n= nhalf;   /* length of the trasform */
  dims[0].is=1;
  dims[0].os=1; 
  int howmany_rank=1; 
  fftw_iodim64 * howmany_dims = malloc(howmany_rank * sizeof(fftw_iodim64)); 
  howmany_dims[0].n=1;
  howmany_dims[0].is=1;
  howmany_dims[0].os=1;
  
  fftwq_plan plan = fftwq_plan_guru64_dft(rank, /*rank=1 means monodimensional vector*/
              dims, /* nhalf */
              howmany_rank,
              howmany_dims,
              fftin,   /*input for INPLACE transforms*/ 
              fftout,  /*output; for INPLACE transforms*/ 
              FFTW_FORWARD, /* -1 because it is the sign of FORWARD transform */
              FFTW_ESTIMATE); /*flags*/
              
  if (plan == NULL) { 
    fprintf(stderr,"plan creation failed\n");
    return 0;
  }
    
  /*  saving  a  generated  wisdom  for  future  runs  */
  if (!fftwq_import_wisdom_from_filename("wisdom-inplace-fftwq-estimate.txt")) {
    fprintf(stderr,  "***  Saving wisdom-inplace-fftwq-estimate.txt \n");
    fftwq_export_wisdom_to_filename("wisdom-inplace-fftwq-estimate.txt"); 
  }
  
  clock_t endplan = clock(); 

  print_time("Plan generation", startplan, endplan);

  return plan;
}

fftwq_complex * mkLGodd(fftwq_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_LGodd) {

  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  __float128 sinvalue;
  __float128 cosvalue;
  __float128 logpi = logq(pi);
  __float128 twist = 0.0Q;
  __float128 angle = 0.0Q; 
  __float128 aux;
  __float128 d;
  long int a;
  ptrdiff_t k;
  
  /* 
   * STEP: mkLGodd
   *  need: fftin mmapped array
   *  provide: file <basename>-LGodd.qd
   */
  
  clock_t startclock = clock();  /* start clock */

  angle = (__float128) pi / nhalf;
  twist = 0.0Q; 
  a = 1; 

  for (k = 0; k <nhalf ; k++){   
     cosvalue = cosq(twist);
     sinvalue = sinq(twist);  
     d = (__float128) a/q ;
     aux = (2.0) * lgammaq( d ) + logq( sinq(pi * d)) - logpi ;  /* lgammaq( d ) - lgammaq( 1.0Q - d );*/
     fftin[k][REAL] = aux * cosvalue;  
     fftin[k][IMAG] = (-1.0Q) * aux * sinvalue; 
     a = (a * primroot) % q;  
     twist = twist + angle;
  }

  clock_t endclock = clock();  /* end clock */  
  print_time("LNGammaOdd computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwq_complex * lgodd = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_LGODD, 'r', fd_LGodd);

  return lgodd;

}

fftwq_complex * fftLGodd(fftwq_complex *fftin, fftwq_complex *fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_LGoddfft) {
  
  /* 
   * STEP: fftLGodd
   *  need: fftin mmapped array with LGodd data
   *  provide: fft of fftin in array fftout
   *  provide: copy of fftout in file <basename>-LGoddfft.qd if basename != NULL
   */

  clock_t startfft = clock(); /* start clock */
  fftwq_execute(plan); 
  clock_t endfft = clock(); /* stop clock */
  print_time("FFT(LGodd)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwq_complex * lgoddfft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTLGODD, 'r', fd_LGoddfft);

  return lgoddfft;

}

fftwq_complex * mkAK(fftwq_complex *fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_AK) {

  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  __float128 sinvalue;
  __float128 cosvalue;
  __float128 twist = 0.0Q;
  __float128 angle = 0.0Q; 
  __float128 c;
  ptrdiff_t a;
  ptrdiff_t k;

  /* 
   * STEP: mkAK
   *  need: fftin mmapped array available for writing
   *  provide: input data AK (in fftin) to compute <basename>-AKfft.qd
   */
  
  clock_t startclock = clock();  /* start clock */  
  angle = (__float128) pi/nhalf;
  twist = 0.0Q; 
  a = 1; 
  for (k = 0; k < nhalf ; k++){  
     c = (__float128) (2*a-q); /*g^k -R g^(k+m) \equiv g^k  -R (q-1)g^k \equiv g^k -R (-g^k) \equiv g^k -R (q -g^k) = 2*g^k -R q */
     cosvalue = cosq(twist);
     sinvalue = - sinq(twist);  /* depends on the sign = -1 of the transform */
     fftin[k][REAL] = c * cosvalue;
     fftin[k][IMAG] = c * sinvalue;
     a = (a * primroot) % q;
     twist += angle;
  }
  clock_t endclock = clock();  /* end clock */
  print_time("AK computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwq_complex * ak = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_AK, 'r', fd_AK);

  return ak;

}

fftwq_complex * fftAK(fftwq_complex *fftin, fftwq_complex *fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_AKfft) {
  
  /* STEP: fftAK 
   *  need: fftin mmapped array with ak start data 
   *  provide: fft in fftout, file <basename>-AKoddfft.qd if basename != NULL
   */

  clock_t startfft = clock();
  fftwq_execute(plan);
  clock_t endfft  = clock();
  print_time("FFT(AK)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  fftwq_complex * akfft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTAK, 'r', fd_AKfft);

  return akfft;

}

struct result oddResult(fftwq_complex *lngammaodd_fft, fftwq_complex *ak_fft, ptrdiff_t nhalf, long int q) {
   
  struct result myresult;
  char bufRe[512];
  
  /* Computing EK(q) - EK^+(q) */
  /* Computing min_{chi\ne chi_0; chi odd} \vert L'/L(1,\chi) \vert =: mq_odd */
  /* Computing max_{chi\ne chi_0; chi odd} \vert L'/L(1,\chi) \vert =: Mq_odd */ 
  
  printf("%s", "Computing the sum and minmax over odd characters"); 
  fflush(stdout);
  
  clock_t startcompodd = clock(); /* start clock */

  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  __float128 oddresult = 0;
  __float128 res;
  __float128 Minimumoddresult;
  __float128 Maximumoddresult;
  __float128 Minimum = FLT128_MAX;/* odd characters */
  __float128 Maximum = 0; 
  __float128 num     = 0;
  __float128 num1    = 0;
  __float128 den     = 0;
  __float128 quot    = 0;   
  __float128 frac    = 0;   
  __float128 frac1   = 0;   
  __float128 corr    = logq(2*pi)+Eulergamma;  
  res = 0;
  oddresult = 0;

  /* computes |(a+ib)/(c+id)|= |a+ib|/|c+id| and the real part of (a+ib)/(c+id); the imaginary part is 0 by def */
  for (long int i=0;  i < nhalf; ++i) { 
    num = lngammaodd_fft[i][REAL] * ak_fft[i][REAL] + lngammaodd_fft[i][IMAG] * ak_fft[i][IMAG];
    den = ak_fft[i][REAL] * ak_fft[i][REAL]+ ak_fft[i][IMAG] * ak_fft[i][IMAG];
    res = res + num/den;
    num1 = lngammaodd_fft[i][IMAG] * ak_fft[i][REAL] - lngammaodd_fft[i][REAL] * ak_fft[i][IMAG];
    frac = corr + q * num / den;
    frac1 = q * num1 / den;
    quot = frac * frac + frac1 * frac1;
    if (quot < Minimum) { Minimum = quot; };
    if (quot > Maximum) { Maximum = quot; }; 
  }
  
  Minimumoddresult = sqrtq(Minimum); /* min for odd characters */
  Maximumoddresult = sqrtq(Maximum); /* max for odd characters */
  oddresult = q * res; /* correction factors for odd characters */

  myresult.res = oddresult;
  myresult.max = Maximumoddresult;
  myresult.min = Minimumoddresult;
  
  clock_t endcompodd = clock(); /* stop clock */
  //printf("\n *** First: oddresult = %32.30Lf, (res = %24.28Lf)\n", oddresult, res);

  printf("\n *** First: oddresult = ");
  { 
  __float128 r = oddresult;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s", bufRe);
  }
 printf(",  (res = ");
  { 
  __float128 r = res;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s", bufRe);
  }
  printf(")\n");
  
  
  print_time("OddResult computation", startcompodd, endcompodd);
  
  fflush(stdout);

  return myresult;
}

fftwq_complex * mkS(fftwq_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, int precision, const char * basename, int * fd_S) {
  
  /* STEP: S 
   *  need: fftin array available for writing
   *  provide: S data in fftin to compute <basename>-Sfft.qd
   *  provide: 
   */
  
  clock_t startcomputeS = clock();
  
  /* inserting the null imaginary parts */   
  long int s_a = 1.0Q; 
  __float128 s_aq = 0.0Q;
  for  (long int k = 0; k < nhalf; k++) { 
    s_aq = (__float128) ((__float128) s_a / (__float128) q);
    //s_aq =  ((__float128) s_a / (__float128) q);
    fftin[k][REAL] = s(s_aq, precision, 0); 
    fftin[k][IMAG] = 0.0Q;
    s_a = (s_a * primroot) % q;
  }
  
  clock_t endcomputeS = clock();
  fflush(stdout);
  print_time("S computation", startcomputeS, endcomputeS);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  fftwq_complex * S = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_S, 'r', fd_S);

  return S;
}

fftwq_complex * fftS(fftwq_complex * fftin, fftwq_complex * fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_Sfft) {
  
  /* STEP: fftS 
   *  need: fftin mmapped array with S data 
   *  provide: file <basename>-Sfft.qd (by copying fftout)
   */
 
  clock_t startfft  = clock();
  fftwq_execute(plan);
  clock_t endfft = clock();
  print_time("Sfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwq_complex * S_fft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTS, 'r', fd_Sfft);

  return S_fft;
}

fftwq_complex * mkLGeven(fftwq_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char *basename, int * fd_LGeven) {
  
  /*
   * STEP: mkLGeven 
   *  need: fftin mmapped array
   *  provide: file <basename>-LGeven.qd
   *
   */
  
  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  //__float128 logpi = logq(pi);
  __float128 d;
  long int a;  
  long int k;

  clock_t startcomp = clock();
  a = 1; 
  for (k = 0; k < nhalf; k++){  
     d = (__float128) a/q ;
     fftin[k][REAL]= - logq(sinq(pi*d)); /* logpi is transformed to 0 by orthogonality of Dirichlet characters*/
     /*fftin[k][REAL]= logpi - logq(sinq(pi*d)); */ /* values of log(gamma(a_k/q) via the reflection formula*/
     //fftin[k][REAL]= lgammaq(d) + lgammaq(1.0Q - d);
     fftin[k][IMAG] = 0.0Q;  
     a = (a * primroot) % q;
  }
  clock_t endcomp = clock();
  print_time("mkLGeven", startcomp, endcomp);
  fflush(stdout);
  
  if (basename == NULL) {
    /* copy of the file not requested, output in fftin */
    return NULL;
  }

  fftwq_complex * LGeven = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_LGEVEN, 'r', fd_LGeven);

  return LGeven;

}

fftwq_complex * fftLGeven(fftwq_complex *fftin, fftwq_complex *fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_LGevenfft) {
  
  /* STEP: fftLGeven 
   *  need: fftin mmapped array with LGeven data
   *  provide: file <basename>-LGevenfft.qd (by renaming fftout)
   */
  
  clock_t startfft = clock();
  fftwq_execute(plan);
  clock_t endfft = clock();
  print_time("LGevenfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwq_complex * LGevenfft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTLGEVEN, 'r', fd_LGevenfft);

  return LGevenfft;

}

struct result evenResult(fftwq_complex *lngamma_evenfft, fftwq_complex *S_fft, ptrdiff_t nhalf, long int q) {

  /* STEP: evenResult */
  
  /* Computing EK^+(q) */
  /* Computing min_{chi\ne chi_0; chi even} \vert L'/L(1,\chi) \vert =: mq_even */
  /* Computing max_{chi\ne chi_0; chi even} \vert L'/L(1,\chi) \vert =: Mq_even */
  /* sum starts from i=1 to avoid the trivial character contribution */

  struct result myresult;
  char bufRe[512];
  
  __float128 evenresult   = 0;
  __float128 res = 0;
  __float128 Minimumevenresult = 0;
  __float128 Maximumevenresult = 0;
  __float128 Minimum = FLT128_MAX; /* even characters */
  __float128 Maximum = 0; 
  __float128 num     = 0;
  __float128 num1    = 0;
  __float128 den     = 0;
  __float128 quot    = 0;   
  __float128 frac    = 0;   
  __float128 frac1   = 0;   
  __float128 corr    = logq(2*pi)+Eulergamma;
  /* ptrdiff_t nquarter = nhalf / 2; */
  /* ptrdiff_t m; */

  res = 0;
  evenresult = 0;

  printf("Computing the sum and minmax over even characters ..."); 
  clock_t startcompeven = clock();

  /* sum starts from i=1 to avoid the trivial character contribution */ 
     
  for (long int i=1;  i < nhalf; ++i) { 
      num = S_fft[i][REAL] * lngamma_evenfft[i][REAL] +  S_fft[i][IMAG]*lngamma_evenfft[i][IMAG];
      den = lngamma_evenfft[i][REAL] *lngamma_evenfft[i][REAL] + lngamma_evenfft[i][IMAG]*lngamma_evenfft[i][IMAG];
      res = res + num / den;
      num1 = S_fft[i][IMAG] * lngamma_evenfft[i][REAL] - S_fft[i][REAL]*lngamma_evenfft[i][IMAG];
      frac = corr + num/(2*den);
      frac1 = num1 / (2*den);
      quot = frac * frac + frac1 * frac1; 
      if (quot < Minimum) { Minimum = quot; };
      if (quot > Maximum) { Maximum = quot; };

  }   
  
  Minimumevenresult = sqrtq(Minimum); /* min for even characters */
  Maximumevenresult = sqrtq(Maximum); /* max for even characters */
  evenresult = res/2.0;

  myresult.res = evenresult;
  myresult.max = Maximumevenresult;
  myresult.min = Minimumevenresult;
  
  clock_t endcompeven = clock();

 // printf("\n *** Second: evenresult = %32.30Lf, (res = %24.28Lf)\n", evenresult, res);
  printf("\n *** Second: evenresult = ");
  { 
  __float128 r = evenresult;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s", bufRe);
  }
 printf(",  (res = ");
  { 
  __float128 r = res;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s", bufRe);
  }
  printf(")\n");
  
  print_time("EvenResult computation", startcompeven, endcompeven);
  
  fflush(stdout);

  return myresult;
}

char bufPrint[512];
void printFloat128(__float128 x) {
  size_t n = quadmath_snprintf(bufPrint, sizeof(bufPrint), "%+-#*.30Qf", 92, x);
  if (n < sizeof(bufPrint))
  printf("%s\n", bufPrint);
  else
  printf("--error--: n = %ld\n", n);
  }

int printResults(long int q, ptrdiff_t nhalf, struct result oddresult, struct result evenresult) {

  char bufRe[512];
 // char bufIm[512];
 // char bufPrint[512];
  
  /** RESULTS **/
  printf("*** RESULTS:\n\n");
  fflush(stdout);
  __float128 EK = Eulergamma + evenresult.res + oddresult.res + (Eulergamma + logq(2*pi)) *(q-2);
  __float128 EK_plus = nhalf*Eulergamma + (nhalf-1)* logq(2*pi) + evenresult.res;
  __float128 EK_diff = oddresult.res + (Eulergamma + logq(2*pi)) * nhalf;
  
  if (EK < 0 || EK_plus < 0 ) {
    FILE * negativeEKfile;
    negativeEKfile = fopen ("../negativeEK.txt","a");
    fprintf (negativeEKfile, "q = %td, ", q);
    //fprintf (negativeEKfile, "EK = %32.30Lf, ", EK); 
     fprintf(negativeEKfile, "EK = "); 
    {
     __float128 r = EK;
     int width = 46;
     int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
     if ((size_t) nquad < sizeof bufRe)
     fprintf (negativeEKfile,"%s", bufRe);
    }
    fprintf(negativeEKfile, ", ");
    
    //fprintf (negativeEKfile, "EK+ = %32.30Lf, ", EK_plus); 
    fprintf(negativeEKfile, "EK+ = "); 
    {
     __float128 r = EK_plus;
     int width = 46;
     int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
     if ((size_t) nquad < sizeof bufRe)
     fprintf (negativeEKfile, "%s", bufRe);
    }
    fprintf(negativeEKfile, ", ");
    
    
    //fprintf (negativeEKfile, "EK-EK+ = %32.30Lf", EK_diff); 
    fprintf(negativeEKfile, "EK-EK+ = "); 
     {
     __float128 r = EK_diff;
     int width = 46;
     int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
     if ((size_t) nquad < sizeof bufRe)
     fprintf (negativeEKfile,"%s", bufRe);
    }
    fprintf(negativeEKfile, ", ");
    fprintf (negativeEKfile, "\n");
    fclose (negativeEKfile);
    }
  

  printf("%s\n", "--------------");
  /* print EK_q */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", EK);
  {
  __float128 r = EK;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }
  
  /* print EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")^+ = ");
 //printf("%32.30Lf\n", EK_plus); 
  {
  __float128 r = EK_plus;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }
  
  /* print EK_q- EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ") - ");
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")^+ = ");
  //printf("%32.30Lf\n", EK_diff); 
  {
  __float128 r = EK_diff;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }

  printf("%s\n", "--------------");

  
  /* print min_q */
  printf("%s", "min_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.min); 
  {
  __float128 r = evenresult.min;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }

  /* print min_q */
  printf("%s", "min_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.min);  
  {
  __float128 r = oddresult.min;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }

  
  /* print min_q */
  printf("%s", "min_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fminq(evenresult.min, oddresult.min)); 
  {
  __float128 r = fminq(evenresult.min, oddresult.min);
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }
   
  
  printf("%s\n", "--------------");
  
  /* print max_q */
  printf("%s", "max_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.max); 
  {
  __float128 r = evenresult.max;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }

  /* print max_q */
  printf("%s", "max_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.max);  
  {
  __float128 r = oddresult.max;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }
  
  /* print min_q */
  printf("%s", "max_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fmaxq(evenresult.max, oddresult.max)); 
  {
  __float128 r = fmaxq(evenresult.max, oddresult.max);
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }  
  
  printf("%s\n", "--------------");  

  return 1;
}

int dostep(const char * step, const char * steps) {
  if (strcasestr(steps, step) != NULL) {
    return 1;
  }
  return 0;
}

int runComputationq(long int q, long int primroot, int precision, const char * basename, const char * steps) {

  ptrdiff_t NUM_POINTS = q-1;
  ptrdiff_t nhalf = NUM_POINTS / 2;

  int fd_fftin;
  int fd_fftout;
  int fd_LGodd;
  int fd_LGoddfft;
  int fd_AK;
  int fd_AKfft;
  int fd_S;
  int fd_Sfft;
  int fd_LGeven;
  int fd_LGevenfft;
  
  fftwq_complex * fftin  = mkfftin(nhalf, basename, &fd_fftin);
  fftwq_complex * fftout = mkfftout(nhalf, basename, &fd_fftout);
  fftwq_complex * lgodd = NULL;
  fftwq_complex * lgoddfft = NULL;
  fftwq_complex * ak = NULL;
  fftwq_complex * akfft = NULL;
  fftwq_complex * S = NULL;
  fftwq_complex * Sfft = NULL;
  fftwq_complex * lgeven = NULL;
  fftwq_complex * lgevenfft = NULL;
  
  struct result oddresult;
  struct result evenresult;
  
  int doall = 0; /* make every step */
  
  if (fftin == NULL || fftout == NULL)
    return 0;

  fftwq_plan plan = mkplan(fftin, fftout, nhalf);

  if (plan == NULL)
    return 0;

  if (dostep(ALL, steps)) {
    doall = 1;
  }
  
  if (doall || dostep(ODDRESULT, steps)) {
  
    if (doall || dostep(MKLGODD, steps)) {
      lgodd     = mkLGodd(fftin, nhalf, q, primroot, basename, &fd_LGodd);       /* qui non servirebbe salvare */
    } else {
      lgodd     = mapfile(N_LGODD, basename, nhalf, &fd_LGodd);
    }
    
    if (doall || dostep(FFTLGODD, steps)) {
      lgoddfft  = fftLGodd(fftin, fftout, plan, nhalf, basename, &fd_LGoddfft);
    } else {
      lgoddfft  = mapfile(N_FFTLGODD, basename, nhalf, &fd_LGoddfft);
    }
    
    if (doall || dostep(MKAK, steps)) {
      ak        = mkAK(fftin, nhalf, q, primroot, basename, &fd_AK);             /* qui non servirebbe salvare */
    } else {
      ak        = mapfile(N_AK, basename, nhalf, &fd_AK);
    }
    
    if (doall || dostep(FFTAK, steps)) {
      akfft     = fftAK(fftin, fftout, plan, nhalf, basename, &fd_AKfft);        /* qui non servirebbe salvare */
    } else {
      akfft     = mapfile(N_FFTAK, basename, nhalf, &fd_AKfft);
    }
    
    oddresult   = oddResult(lgoddfft, akfft, nhalf, q);

    writeResult(oddresult, basename, N_ODDRESULT);
    
    unmap_File2Array(lgodd, nhalf, fd_LGodd);
    unmap_File2Array(lgoddfft, nhalf, fd_LGoddfft);
    unmap_File2Array(ak, nhalf, fd_AK);
    unmap_File2Array(akfft, nhalf, fd_AKfft);

  } else {

    oddresult = readResult(basename, N_ODDRESULT);

  }


  if (doall || dostep(EVENRESULT, steps)) {
  
    if (doall || dostep(MKS, steps)) {
      S         = mkS(fftin, nhalf, q, primroot, precision, basename, &fd_S);    /* qui non servirebbe salvare */
    } else {
      S         = mapfile(N_S, basename, nhalf, &fd_S);
    }
    
    if (doall || dostep(FFTS, steps)) {
      Sfft      = fftS(fftin, fftout, plan, nhalf, basename, &fd_Sfft);
    } else {
      Sfft      = mapfile(N_FFTS, basename, nhalf, &fd_Sfft);
    }
    
    if (doall || dostep(MKLGEVEN, steps)) {
      lgeven    = mkLGeven(fftin, nhalf, q, primroot, basename, &fd_LGeven);      /* qui non servirebbe salvare */
    } else {
      lgeven    = mapfile(N_LGEVEN, basename, nhalf, &fd_LGeven);
    }
    
    if (doall || dostep(FFTLGEVEN, steps)) {
      lgevenfft = fftLGeven(fftin, fftout, plan, nhalf, basename, &fd_LGevenfft); /* qui non servirebbe salvare */
    } else {
      lgevenfft = mapfile(N_FFTLGEVEN, basename, nhalf, &fd_LGevenfft);
    }
    
    evenresult  = evenResult(lgevenfft, Sfft, nhalf, q);

    writeResult(evenresult, basename, N_EVENRESULT);
    
    unmap_File2Array(S, nhalf, fd_S);
    unmap_File2Array(Sfft, nhalf, fd_Sfft);
    unmap_File2Array(lgeven, nhalf, fd_LGeven);
    unmap_File2Array(lgevenfft, nhalf, fd_LGevenfft);
    
  } else {

    evenresult = readResult(basename, N_EVENRESULT);

  }

  printResults(q, nhalf, oddresult, evenresult);

  fftwq_destroy_plan(plan);
  fftwq_cleanup();

  return 0;
}

